#include "bzio.h"
#include "mem.h"
#include "log.h"
#include <bzlib.h>

static void * bzalloc(void *unused, int n, int m)
 {
  return malloc(n*m);
 }

static void bzfree(void *unused, void *data)
 {
  free(data);
 }

typedef struct
 {
  /* This chunk should match the bzf structure. */
  FILE * file;
  
  bz_stream stream;
  char data[4096];
 } bzf_info;

bzf * bz_OpenWrite(const char *filename)
 {
  bzf_info *data = (bzf_info *)malloc(sizeof(bzf_info));
  
  if(!data)
   return 0;
  
  data->stream.next_in = 0;
  data->stream.avail_in = 0;
  data->stream.next_out = 0;
  data->stream.avail_out = 0;
  data->stream.bzalloc = &bzalloc;
  data->stream.bzfree = &bzfree;
  data->stream.opaque = 0;
  
  if(BZ_OK != BZ2_bzCompressInit(&data->stream, 9, 0, 0))
   {
    free(data);
    return 0;
   }
  
  if(!(data->file = fopen(filename, "wb")))
   {
    BZ2_bzCompressEnd(&data->stream);
    free(data);
    return 0;
   }
  
  return (bzf *)data;
 }

bzf * bz_OpenRead(const char *filename)
 {
  bzf_info *data = (bzf_info *)malloc(sizeof(bzf_info));
  
  if(!data)
   return 0;
  
  data->stream.next_in = 0;
  data->stream.avail_in = 0;
  data->stream.next_out = 0;
  data->stream.avail_out = 0;
  data->stream.bzalloc = &bzalloc;
  data->stream.bzfree = &bzfree;
  data->stream.opaque = 0;
  
  if(BZ_OK != BZ2_bzDecompressInit(&data->stream, 0, 0))
   {
    free(data);
    return 0;
   }
  
  if(!(data->file = fopen(filename, "rb")))
   {
    BZ2_bzDecompressEnd(&data->stream);
    free(data);
    return 0;
   }
  
  return (bzf *)data;
 }

int bz_Read(bzf *_data, void *buf, unsigned int len)
 {
  bzf_info *data = (bzf_info *)_data;
  
  if(data && data->file)
   {
    data->stream.next_out = (char *)buf;
    data->stream.avail_out = len;
    
    while(data->stream.avail_out)
     {
      if(!data->stream.avail_in)
       {
        int read = fread(data->data, 1, sizeof data->data, data->file);
        
        if(read == 0)
         {
          logMessage(Warning, "Error reading data from stream.");
          fclose(data->file);
          data->file = 0;
          return 0;
         }
        
        data->stream.next_in = data->data;
        data->stream.avail_in = read;
       }
      
      switch(BZ2_bzDecompress(&data->stream))
       {
        case BZ_OK:
        break;
        
        case BZ_STREAM_END:
         if(!data->stream.avail_out)
          return 1;
         
         logMessage(Warning, "Reached end of compressed stream, but more data was requested.");
         fclose(data->file);
         data->file = 0;
        return 0;
        
        default:
        case BZ_PARAM_ERROR:
        case BZ_DATA_ERROR:
        case BZ_DATA_ERROR_MAGIC:
        case BZ_MEM_ERROR:
         logMessage(Warning, "Error decompressing data from stream.");
         fclose(data->file);
         data->file = 0;
        return 0;
       }
     }
    
    return 1;
   } else return 0;
 }

int bz_Write(bzf *_data, const void *buf, unsigned int len)
 {
  bzf_info *data = (bzf_info *)_data;
  
  if(data && data->file)
   {
    data->stream.next_in = (char *)buf;
    
    if((data->stream.avail_in = len)) while(data->stream.avail_in)
     {
      const char *out, *out_end;
      
      data->stream.next_out = data->data;
      data->stream.avail_out = sizeof data->data;
      
      if(BZ_RUN_OK != BZ2_bzCompress(&data->stream, BZ_RUN))
       {
        logMessage(Warning, "Error compressing data for stream.");
        fclose(data->file);
        data->file = 0;
        return 0;
       }
      
      out = data->data;
      out_end = data->stream.next_out;
      
      while(out != out_end)
       {
        int written = fwrite(out, 1, out_end-out, data->file);
        
        if(written == 0)
         {
          logMessage(Warning, "Error writing compressed data to file.");
          fclose(data->file);
          data->file = 0;
          return 0;
         }
        
        out += written;
       }
     }
    
    return 1;
   }
  else return 0;
 }

void bz_CloseWrite(bzf *_data)
 {
  bzf_info *data = (bzf_info *)_data;
  
  if(data)
   {
    if(data->file)
     {
      int state;
      
      data->stream.next_in = 0;
      data->stream.avail_in = 0;
      
      do
       {
        const char *out, *out_end;
        
        data->stream.next_out = data->data;
        data->stream.avail_out = sizeof data->data;
        
        state = BZ2_bzCompress(&data->stream, BZ_FINISH);
        
        if(state != BZ_FINISH_OK && state != BZ_STREAM_END)
         {
          logMessage(Warning, "Error compressing data for stream while closing.");
          BZ2_bzCompressEnd(&data->stream);
          fclose(data->file);
          free(data);
          return;
         }
        
        out = data->data;
        out_end = data->stream.next_out;
        while(out != out_end)
         {
          int written = fwrite(out, 1, out_end-out, data->file);
          
          if(written == 0)
           {
            logMessage(Warning, "Error writing compressed data to file while closing.");
            BZ2_bzCompressEnd(&data->stream);
            fclose(data->file);
            free(data);
            return;
           }
          
          out += written;
         }
       } while(state == BZ_FINISH_OK);
      
      if(fflush(data->file))
       {
        logMessage(Warning, "Error flushing file stream while closing.");
        BZ2_bzCompressEnd(&data->stream);
        fclose(data->file);
        free(data);
        return;
       }
     }
    
    BZ2_bzCompressEnd(&data->stream);
    
    if(data->file) fclose(data->file);
    free(data);
   }
 }

void bz_CloseRead(bzf *_data)
 {
  bzf_info *data = (bzf_info *)_data;
  
  if(data)
   {
    BZ2_bzDecompressEnd(&data->stream);
    
    if(data->file) fclose(data->file);
    free(data);
   }
 }

int bz_Flush(bzf *_data)
 {
  bzf_info *data = (bzf_info *)_data;
  
  if(data && data->file)
   {
    int state;
    
    data->stream.next_in = 0;
    data->stream.avail_in = 0;
    
    do
     {
      const char *out, *out_end;
      
      data->stream.next_out = data->data;
      data->stream.avail_out = sizeof data->data;
      
      state = BZ2_bzCompress(&data->stream, BZ_FLUSH);
      
      if(state != BZ_FLUSH_OK && state != BZ_RUN_OK)
       {
        logMessage(Warning, "Error compressing data for stream while flushing.");
        fclose(data->file);
        data->file = 0;
        return 0;
       }
      
      out = data->data;
      out_end = data->stream.next_out;
      while(out != out_end)
       {
        int written = fwrite(out, 1, out_end-out, data->file);
        
        if(written == 0)
         {
          logMessage(Warning, "Error writing compressed data to file while flushing.");
          fclose(data->file);
          data->file = 0;
          return 0;
         }
        
        out += written;
       }
     } while(state == BZ_FLUSH_OK);
    
    if(fflush(data->file))
     {
      logMessage(Warning, "Error flushing file stream.");
      fclose(data->file);
      data->file = 0;
      return 0;
     }
    
    return 1;
   }
  else return 0;
 }
