#ifndef CANVAS_H
#define CANVAS_H

#include <math.h>
#include <stdint.h>
#include <assert.h>

#include "region.h"

typedef struct
 {
  float x, y;
 } CanvasPoint;

typedef struct
 {
  /* The width and height of the canvas, in pixels.
     Must be multiples of 256. The OpenGL canvas renderer demands it.
     Maybe later this requirement could be changed to n*256+2^m where n and m are positive integers. */
  uint16_t width, height;
  
  /* How many bytes to add to a pixel address to move one row down. Will always be width*3. */
  uint32_t pitch;
  
  /* Pointer to the pixel data, specifically the start of the top/left pixel. Not managed by
     the canvas itself. */
  uint8_t *data;
  
  CanvasPoint *points_begin, *points_end, *points_alloc_end;
  
  /* Marks the regions that have changed. */
  DirtyRectangles dirt;
 } Canvas;

extern int initCanvas(Canvas *canvas, uint16_t width, uint16_t height, uint8_t *data);
extern void uninitCanvas(Canvas *canvas);

extern Canvas *createCanvas(uint16_t width, uint16_t height);
extern void destroyCanvas(Canvas *canvas);

/* Function for testing. */
extern void dumpCanvasPoints(Canvas *canvas, uint8_t r, uint8_t g, uint8_t b, uint8_t a);

extern void clearCanvasPoints(Canvas *canvas);
extern int pushCanvasPoint(Canvas *canvas, CanvasPoint p);
extern int pushCanvasLine(Canvas *canvas, CanvasPoint start, CanvasPoint end, float *remainder);

static inline const unsigned char *getReadPixelUnsafe(const Canvas *canvas, uint16_t x, uint16_t y)
 {
  assert(canvas);
  assert(x < canvas->width);
  assert(y < canvas->height);
  
  return canvas->data+x*3+y*canvas->pitch;
 }

static inline unsigned char *getWritePixelUnsafe(Canvas *canvas, uint16_t x, uint16_t y)
 {
  assert(canvas);
  assert(x < canvas->width);
  assert(y < canvas->height);
  
  return canvas->data+x*3+y*canvas->pitch;
 }

static inline void writePixel(uint8_t *p, uint8_t r, uint8_t g, uint8_t b)
 {
  *p = r; ++p;
  *p = g; ++p;
  *p = b;
 }

static inline void setPixelUnsafe(Canvas *canvas, uint16_t x, uint16_t y, uint8_t r, uint8_t g, uint8_t b)
 {
  writePixel(getWritePixelUnsafe(canvas, x, y), r, g, b);
 }

static inline void writeAlphaPixel(uint8_t *p, uint8_t r, uint8_t g, uint8_t b, uint8_t a)
 {
  uint8_t ia = ~a;
  *p = (*p*ia+r*a)/255; ++p;
  *p = (*p*ia+g*a)/255; ++p;
  *p = (*p*ia+b*a)/255;
 }

static inline void setAlphaPixelUnsafe(Canvas *canvas, uint16_t x, uint16_t y, uint8_t r, uint8_t g, uint8_t b, uint8_t a)
 {
  writeAlphaPixel(getWritePixelUnsafe(canvas, x, y), r, g, b, a);
 }

static inline void setPixel(Canvas *canvas, uint16_t x, uint16_t y, uint8_t r, uint8_t g, uint8_t b)
 {
  if(x < canvas->width && y < canvas->height) setPixelUnsafe(canvas, x, y, r, g, b);
 }

static inline void setAlphaPixel(Canvas *canvas, uint16_t x, uint16_t y, uint8_t r, uint8_t g, uint8_t b, uint8_t a)
 {
  if(x < canvas->width && y < canvas->height) setAlphaPixelUnsafe(canvas, x, y, r, g, b, a);
 }

extern void drawCanvasPoint(Canvas *canvas, CanvasPoint p, uint8_t r, uint8_t g, uint8_t b, uint8_t a);
extern float getCanvasColour(const Canvas *canvas, CanvasPoint p, float *pixel);

#endif
