#include <math.h>

#include "mem.h"
#include "assert.h"
#include "draw.h"
#include "rng.h"
#include "config.h"

#include <math.h>

#define SQR(X) ((X)*(X))

void convertColourToHSV(Colour c, HSV *h)
 {
  assert(h);
  
  rgb_to_hsv(c.r, c.g, c.b, &h->h, &h->s, &h->v);
 }

void convertHSVToColour(HSV h, Colour *c)
 {
  int r, g, b;
  
  assert(c);
  
  hsv_to_rgb(h.h, h.s, h.v, &r, &g, &b);
  c->r = r;
  c->g = g;
  c->b = b;
 }

void getSampleColour(BITMAP *surface, int16_t x, int16_t y, Colour *c)
 {
  float total;
  int16_t u, v;
  
  assert(surface);
  assert(c);
  
  total = 0;
  
  for(v = -4; v <= 4; ++v)
   for(u = -4; u <= 4; ++u)
    if(x+u >= 0 && y+v >= 0 && x+u < surface->w && y+v < surface->h)
     {
      int colour = _getpixel32(surface, x+u, y+v);
      c->r += getr32(colour);
      c->g += getg32(colour);
      c->b += getb32(colour);
      ++total;
     }
  
  if(total > 0)
   {
    int iter;
    Colour average;
    average.r = c->r/total;
    average.g = c->g/total;
    average.b = c->b/total;
    
    for(iter = 0; iter < 150; ++iter)
     {
      total = 0.0;
      c->r = c->g = c->b = 0.0;
      
      for(v = -4; v <= 4; ++v)
       for(u = -4; u <= 4; ++u)
        if(x+u >= 0 && y+v >= 0 && x+u < surface->w && y+v < surface->h)
         {
          int colour = _getpixel32(surface, x+u, y+v);
          float dist = 0.1/(0.1+sqrt(SQR((getr32(colour)-average.r)/255.0)+
                                     SQR((getg32(colour)-average.g)/255.0)+
                                     SQR((getb32(colour)-average.b)/255.0)));
          
          c->r += getr32(colour)*dist;
          c->g += getg32(colour)*dist;
          c->b += getb32(colour)*dist;
          total += dist;
         }
      
      average.r = c->r/total;
      average.g = c->g/total;
      average.b = c->b/total;
     }
    
    c->r = average.r;
    c->g = average.g;
    c->b = average.b;
   }
  else
   c->r = c->g = c->b = 127.5;
 }

inline void alphaPixelUnsafe(BITMAP *bmp, int x, int y, Colour c, float alpha)
 {
  assert(isfinite(alpha));
  assert(alpha >= 0.0 && alpha <= 1.0);
  assert(bmp);
  assert(x >= 0 && x < bmp->w);
  assert(y >= 0 && y < bmp->h);
  
  if(alpha > 0.999)
   _putpixel32(bmp, x, y, makecol32(c.r, c.g, c.b));
  else
   {
    int colour = _getpixel32(bmp, x, y);
    c.r *= alpha;
    c.g *= alpha;
    c.b *= alpha;
    alpha = 1.0-alpha;
    _putpixel32(bmp, x, y, makecol32(c.r+getr32(colour)*alpha+0.5, c.g+getg32(colour)*alpha+0.5, c.b+getb32(colour)*alpha+0.5));
   }
 }

inline void alphaPixel(BITMAP *bmp, int x, int y, Colour c, float alpha)
 {
  assert(isfinite(alpha));
  assert(alpha >= 0.0 && alpha <= 1.0);
  assert(bmp);
  
  if(alpha > 0.005 && x >= 0 && y >= 0 && x < bmp->w && y < bmp->h)
   alphaPixelUnsafe(bmp, x, y, c, alpha);
 }

void alphaHLine(BITMAP *bmp, int x1, int y, int x2, Colour c, float alpha)
 {
  assert(isfinite(x1) && isfinite(y) && isfinite(x2));
  assert(isfinite(alpha));
  assert(alpha >= 0.0 && alpha <= 1.0);
  assert(bmp);
  
  if(alpha > 0.005 && y >= 0 && y < bmp->h)
   {
    if(x1 < 0) x1 = 0;
    if(x2 >= bmp->w) x2 = bmp->w - 1;
    
    if(alpha > 0.999)
     {
      int col = makecol32(c.r, c.g, c.b);
      
      for(; x1 <= x2; ++x1)
       _putpixel32(bmp, x1, y, col);
     }
    else
     {
      c.r *= alpha;
      c.g *= alpha;
      c.b *= alpha;
      alpha = 1.0-alpha; 
      
      for(; x1 <= x2; ++x1)
       {
        int colour = _getpixel32(bmp, x1, y);
        _putpixel32(bmp, x1, y, makecol32(c.r+getr32(colour)*alpha+0.5, c.g+getg32(colour)*alpha+0.5, c.b+getb32(colour)*alpha+0.5));
       }
     }
   }
 }

inline void drawPointUnsafe(BITMAP *bmp, float x, float y, Colour c, float alpha)
 {
  int ix, iy;
  float fx, ifx, fy, ify;
  
  assert(isfinite(x) && isfinite(y));
  assert(isfinite(alpha));
  assert(alpha >= 0.0 && alpha <= 1.0);
  assert(bmp);
  assert(x >= 0 && x+1.0001 < bmp->w);
  assert(y >= 0 && y+1.0001 < bmp->h);
  
  ix = floor(x);
  iy = floor(y);
  
  assert(ix >= 0 && ix+1 < bmp->w);
  assert(iy >= 0 && iy+1 < bmp->h);
  
  fx = x-ix;
  ifx = 1.0-fx;
  fy = y-iy;
  ify = 1.0-fy;
  
  alphaPixelUnsafe(bmp, ix, iy, c, ifx*ify*alpha);
  alphaPixelUnsafe(bmp, ix+1, iy, c, fx*ify*alpha);
  alphaPixelUnsafe(bmp, ix+1, iy+1, c, fx*fy*alpha);
  alphaPixelUnsafe(bmp, ix, iy+1, c, ifx*fy*alpha);
 }

inline void drawPoint(BITMAP *bmp, float x, float y, Colour c, float alpha)
 {
  assert(isfinite(x) && isfinite(y));
  assert(isfinite(alpha));
  assert(alpha >= 0.0 && alpha <= 1.0);
  assert(bmp);
  
  if(x >= 0 && y >= 0 && x+1.0001 < bmp->w && y+1.0001 < bmp->h)
   drawPointUnsafe(bmp, x, y, c, alpha);
 }

void drawPaper(BITMAP *bmp)
 {
  Colour c = {232, 229, 227}, s = {186, 178, 171}, h = {255, 255, 255};
  float u, v;
  
  assert(bmp);
  
  clear_to_color(bmp, makecol32(c.r, c.g, c.b));
  
  RNGPush(0);
  
  for(v = 0; v < bmp->h; v += 2.5)
   for(u = 0; u < bmp->w; ++u)
    {
     drawPoint(bmp, u+(fRNG()-0.5)*1.4-0.5, v+(fRNG()-0.5)*1.4, s, 0.2);
     drawPoint(bmp, u+(fRNG()-0.5)*1.4-0.5, v+(fRNG()-0.5)*1.4+2.3*0.3, h, 0.1);
    }
  
  for(v = 0; v < bmp->h; ++v)
   for(u = 0; u < bmp->w; u += 2.5)
    {
     drawPoint(bmp, u+(fRNG()-0.5)*1.4+2.3*0.3, v+(fRNG()-0.5)*1.4, h, 0.2);
     drawPoint(bmp, u+(fRNG()-0.5)*1.4-0.5, v+(fRNG()-0.5)*1.4, s, 0.1);
    }
  
  RNGPop();
 }

inline float drawLineRemainder(BITMAP *bmp, float x1, float y1, float x2, float y2, Colour c, float alpha, float remainder)
 {
  float distance, dx, dy;
  
  assert(isfinite(x1) && isfinite(y1));
  assert(isfinite(x2) && isfinite(y2));
  assert(isfinite(remainder));
  assert(isfinite(alpha));
  assert(alpha >= 0.0 && alpha <= 1.0);
  assert(remainder >= 0.0);
  assert(bmp);
  
  distance = sqrt((x2-x1)*(x2-x1)+(y2-y1)*(y2-y1));
  
  if(distance < remainder)
   return remainder - distance;
  else
   {
    float idistance = 1.0/distance;
    dx = (x2-x1)*idistance;
    dy = (y2-y1)*idistance;
    drawPoint(bmp, x1 += dx*remainder, y1 += dy*remainder, c, alpha);
    distance -= remainder;
   }
  
  assert(isfinite(dx) && isfinite(dy));
  
  while(1)
   {
    if(distance < 1.0)
     return 1.0 - distance;
    else
     {
      --distance;
      drawPoint(bmp, x1 += dx, y1 += dy, c, alpha);
     }
   }
 }

inline void drawLine(BITMAP *bmp, float x1, float y1, float x2, float y2, Colour c, float alpha)
 {
  assert(isfinite(x1) && isfinite(y1));
  assert(isfinite(x2) && isfinite(y2));
  assert(isfinite(alpha));
  assert(alpha >= 0.0 && alpha <= 1.0);
  assert(bmp);
  
  drawLineRemainder(bmp, x1, y1, x2, y2, c, alpha, 0.0);
 }

int getColour(BITMAP *surface, float x, float y, Colour *c)
 {
  int ix, iy;
  
  assert(isfinite(x) && isfinite(y));
  assert(surface);
  assert(c);
  
  ix = floor(x);
  iy = floor(y);
  
  if(ix >= 0 && iy >= 0 && ix+1.0001 < surface->w && iy+1.0001 < surface->h)
   {
    float fx = x-ix, ifx = 1.0-fx, fy = y-iy, ify = 1.0-fy;
    int colour = _getpixel32(surface, ix, iy);
    
    c->r = getr32(colour)*(ifx*ify);
    c->g = getg32(colour)*(ifx*ify);
    c->b = getb32(colour)*(ifx*ify);
    
    colour = _getpixel32(surface, ix+1, iy);
    c->r += getr32(colour)*(fx*ify),
    c->g += getg32(colour)*(fx*ify),
    c->b += getb32(colour)*(fx*ify);
    
    colour = _getpixel32(surface, ix, iy+1);
    c->r += getr32(colour)*(ifx*fy),
    c->g += getg32(colour)*(ifx*fy),
    c->b += getb32(colour)*(ifx*fy);
    
    colour = _getpixel32(surface, ix+1, iy+1);
    c->r += getr32(colour)*(fx*fy),
    c->g += getg32(colour)*(fx*fy),
    c->b += getb32(colour)*(fx*fy);
    
    return 1;
   }
  else return 0;
 }

inline void drawCircle(BITMAP *bmp, float x, float y, Colour c, float alpha, float radius)
 {
  float fx, fy, temp, dist;
  int cr, v;
  
  assert(isfinite(x) && isfinite(y) && isfinite(alpha) && isfinite(radius));
  assert(radius > 0.0);
  assert(bmp);
  
  fx = x-floor(x);
  fy = y-floor(y);
  cr = floor(radius+0.5);
  
  /* TODO: Need a special case for v=1-cr and v=cr; those rows aren't getting anti-aliased.
     Good enough for now; only noticable on large circles. */
  
  for(v = 1-cr; v <= cr; ++v)
   {
    int sr;
    
    if((temp = radius*radius-SQR(v-fy)) <= 0)
     continue;
    
    sr = floor(sqrt(temp)+0.5);
    
    alphaHLine(bmp, x-sr+2, y+v, x+sr-1, c, alpha);
    
    if((dist = SQR(1-sr-fx)+SQR(v-fy)) > 0)
     {
      if((dist = sqrt(dist)-radius) < -1.0) alphaPixel(bmp, x-sr+1, y+v, c, alpha);
      else if(dist < 0.0) alphaPixel(bmp, x-sr+1, y+v, c, -dist*alpha);
     }
    
    if(sr && (dist = SQR(sr-fx)+SQR(v-fy)) > 0)
     {
      if((dist = sqrt(dist)-radius) < -1.0) alphaPixel(bmp, x+sr, y+v, c, alpha);
      else if(dist < 0.0) alphaPixel(bmp, x+sr, y+v, c, -dist*alpha);
     }
   }
 }

#define MAX_SPLINE_POINTS 8192
float spline_x[MAX_SPLINE_POINTS], spline_y[MAX_SPLINE_POINTS];
unsigned int spline_points = 0;

float calcSplinePoints(float x1, float y1, float x2, float y2, float x3, float y3, float x4, float y4, float spacing, float remainder)
 {
  static const float threshold = 0.01;
  float p = 0, ip, old_x = x1, old_y = y1, new_x, new_y, diff,
        remainder_max, remainder_min, spacing_sqrd, spacing_sqrd_min, spacing_sqrd_max;
  
  assert(isfinite(x1) && isfinite(y1));
  assert(isfinite(x2) && isfinite(y2));
  assert(isfinite(x3) && isfinite(y3));
  assert(isfinite(x4) && isfinite(y4));
  assert(isfinite(spacing) && isfinite(remainder));
  assert(remainder >= 0.0);
  assert(spacing > 0.01);
  
  spline_points = 0;
  
  remainder_max = SQR(remainder+threshold),
  remainder_min = SQR(remainder-threshold),
  spacing_sqrd =  SQR(spacing),
  spacing_sqrd_min = SQR(spacing-threshold),
  spacing_sqrd_max = SQR(spacing+threshold);
  remainder = SQR(remainder);
  
  while(1)
   {
    float min = p, max = 1.0, r;
    
    if((r = SQR(x4-old_x)+SQR(y4-old_y)) < remainder)
     return sqrt(remainder) - sqrt(r);
    
    while(1)
     {
      p = (min+max)*0.5;
      ip = 1.0-p;
      
      new_x = x1*ip*ip*ip + x2*3.0*ip*ip*p + x3*3.0*ip*p*p + x4*p*p*p;
      new_y = y1*ip*ip*ip + y2*3.0*ip*ip*p + y3*3.0*ip*p*p + y4*p*p*p;
      
      assert(isfinite(p) && isfinite(ip));
      assert(isfinite(new_x) && isfinite(new_y));
      
      if((diff = SQR(old_x-new_x)+SQR(old_y-new_y)) > remainder_max)
       max = p;
      else if(diff < remainder_min && max-min >= 0.0001)
       min = p;
      else break;
     }
    
    if(spline_points != MAX_SPLINE_POINTS)
     {
      spline_x[spline_points] = new_x;
      spline_y[spline_points] = new_y;
      ++spline_points;
      
      remainder = spacing_sqrd;
      remainder_min = spacing_sqrd_min;
      remainder_max = spacing_sqrd_max;
      old_x = new_x;
      old_y = new_y;
     }
    else return 0.0;
   }
 }

float calcLinePoints(float x1, float y1, float x2, float y2, float spacing, float remainder)
 {
  float distance, dx, dy;
  
  assert(isfinite(x1) && isfinite(y1) && isfinite(x2) && isfinite(y2));
  assert(isfinite(spacing) && isfinite(remainder));
  assert(remainder >= 0.0);
  assert(spacing > 0.01);
  
  distance = sqrt((x2-x1)*(x2-x1)+(y2-y1)*(y2-y1));
  
  if(distance < remainder || distance < 0.000000001)
   {
    spline_points = 0;
    return remainder - distance;
   }
  
  dx = (x2-x1)/distance;
  dy = (y2-y1)/distance;
  
  assert(isfinite(dx) && isfinite(dy));
  
  spline_x[0] = (x1 += dx*remainder);
  spline_y[0] = (y1 += dy*remainder);
  spline_points = 1;
  distance -= remainder;
  
  while(1)
   {
    if(distance < spacing)
     return spacing - distance;
    else if(spline_points != MAX_SPLINE_POINTS)
     {
      distance -= spacing;
      spline_x[spline_points] = (x1 += dx*spacing);
      spline_y[spline_points] = (y1 += dy*spacing);
      ++spline_points;
     }
    else return 0.0;
   }
 }

float calcSplineChunk(const int16_t *x, const int16_t *y, uint32_t points, uint32_t index, float spacing, float remainder)
 {
  assert(isfinite(remainder));
  
  assert(x);
  assert(y);
  
  if(points > 2)
   {
    float x1, y1, x2, y2, x3, y3, x4, y4, d1, d2;
    
    if(index >= points - 1)
     {
      spline_points = 0;
      return remainder;
     }
    
    x1 = x[index];
    x4 = x[index+1];
    y1 = y[index];
    y4 = y[index+1];
    
    if(index == 0)
     {
      x2 = x[index+1]-x[index];
      y2 = y[index+1]-y[index];
     }
    else
     {
      x2 = x[index+1]-x[index-1];
      y2 = y[index+1]-y[index-1];
     }
    
    if(index == points - 2)
     {
      x3 = x[index]-x[index+1];
      y3 = y[index]-y[index+1];
     }
    else
     {
      x3 = x[index]-x[index+2];
      y3 = y[index]-y[index+2];
     }
    
    d1 = sqrt(SQR(x2)+SQR(y2));
    d2 = sqrt(SQR(x3)+SQR(y3));
    
    if(d2 < 0.001 || d1 < 0.001)
     return calcLinePoints(x1, y1, x4, y4, spacing, remainder);
    else if(d2 > d1)
     return calcSplinePoints(x1, y1, x1+x2*0.25, y1+y2*0.25, x4+x3/d2*d1*0.25, y4+y3/d2*d1*0.25, x4, y4, spacing, remainder);
    else
     return calcSplinePoints(x1, y1, x1+x2/d1*d2*0.25, y1+y2/d1*d2*0.25, x4+x3*0.25, y4+y3*0.25, x4, y4, spacing, remainder);
   }
  else if(index == 0)
   {
    if(points == 2)
     return calcLinePoints(*x, *y, x[1], y[1], spacing, remainder);
    else if(points == 1 && remainder < 0.0001)
     { 
      spline_x[0] = *x;
      spline_x[1] = *y;
      spline_points = 1;
      return spacing;
     }
    else return remainder;
   }
  else
   {
    spline_points = 0;
    return remainder;
   }
 }

float drawPreviewLine(BITMAP *surface, Colour c, const int16_t *x, const int16_t *y, uint32_t points, uint32_t first, uint32_t last, float remainder, Region *r)
 {
  char first_assign = 1;
  uint32_t u = 0, v = first;
  float offx, offy;
  
  assert(isfinite(remainder));
  assert(surface);
  assert(x && y);
  assert(first < last);
  assert(last >= first && last <= points);
  
  offx = (surface->w-buffer_w)*0.5;
  offy = (surface->h-buffer_h)*0.5;
  
  for(; v < last; ++v)
   {
    remainder = calcSplineChunk(x, y, points, v, 1.0, remainder);
    
    for(u = 0; u < spline_points; ++u)
     {
      if(spline_x[u] >= 0 && spline_y[u] >= 0 && spline_x[u] < buffer_w-1 && spline_y[u] < buffer_h-1)
       {
        drawPoint(surface, spline_x[u]+offx, spline_y[u]+offy, c, 1.0);
        
        if(r)
         {
          if(first_assign)
           {
            r->x1 = floor(spline_x[u]);
            r->y1 = floor(spline_y[u]);
            r->x2 = ceil(spline_x[u])+2;
            r->y2 = ceil(spline_y[u])+2;
            first_assign = 0;
           }
          else
           {
            int16_t temp;
            
            if((temp = floor(spline_x[u])) < r->x1)
             r->x1 = temp;
            
            if((temp = floor(spline_y[u])) < r->y1)
             r->y1 = temp;
            
            if((temp = ceil(spline_x[u])+2) > r->x2)
             r->x2 = temp;
            
            if((temp = ceil(spline_y[u])+2) > r->y2)
             r->y2 = temp;
           }
         }
       }
     }
   }
  
  if(r)
   {
    if(first_assign)
     {
      r->x1 = r->y1 = r->x2 = r->y2 = 0;
     }
    else
     {
      r->x1 += offx;
      r->y1 += offy;
      r->x2 += offx;
      r->y2 += offy;
     }
   }
  
  return remainder;
 }

void drawPen(BITMAP *surface, Colour c, const int16_t *x, const int16_t *y, uint32_t points)
 {
  float remainder = 0.0;
  uint32_t u, v;
  
  logMessage(Debug, "drawPen invoked with %d points.", points);
  
  assert(surface);
  assert(x);
  assert(y);
  
  for(v = 0; v < points; ++v)
   {
    remainder = calcSplineChunk(x, y, points, v, 1.0, remainder);
    
    for(u = 0; u < spline_points; ++u)
     drawPoint(surface, spline_x[u], spline_y[u], c, 1.0);
   }
 }

const float chalk_min_radius = 1.5, chalk_max_radius = 12.0;

void drawChalk(BITMAP *surface, Colour c, float radius, const int16_t *x, const int16_t *y, uint32_t points)
 {
  float remainder = 0.0, area = 0;
  uint32_t u, v;
  
  logMessage(Debug, "drawChalk invoked with %d points.", points);
  
  assert(isfinite(radius));
  
  assert(surface);
  assert(x && y);
  
  if(radius < chalk_min_radius) radius = chalk_min_radius;
  else if(radius > chalk_max_radius) radius = chalk_max_radius;
  
  for(v = 0; v < points; ++v)
   {
    remainder = calcSplineChunk(x, y, points, v, 1.0, remainder);
    
    for(u = 0; u < spline_points; ++u)
     {
      for(area = 0; area <= radius*radius*3.1415926535897932384626433832795*1.5;)
       {
        float _x, _y, dist, offset = fRNG()*30.0-15.0, r;
        Colour t;
        t.r = c.r+offset;
        t.g = c.g+offset;
        t.b = c.b+offset;
        
        if(offset > 0)
         {
          if(t.r > 255.0) t.r = 255.0;
          if(t.g > 255.0) t.g = 255.0;
          if(t.b > 255.0) t.b = 255.0;
         }
        else
         {
          if(t.r < 0.0) t.r = 0.0;
          if(t.g < 0.0) t.g = 0.0;
          if(t.b < 0.0) t.b = 0.0;
         }
        
        do
         {
          _x = (fRNG()-0.5)*2.0*radius;
          _y = (fRNG()-0.5)*2.0*radius;
         } while((dist = sqrt(SQR(_x)+SQR(_y))) > radius);
        
        if((r = (radius-dist)*(fRNG()*0.4+0.2)) < 0.2)
         r = 0.2;
        
        drawCircle(surface, spline_x[u]+_x, spline_y[u]+_y, t, 0.5, r);
        area += (r*r*3.1415926535897932384626433832795)*(r*r*3.1415926535897932384626433832795);
       }
     }
   }
 }


typedef struct {float x, y, lx, ly, remain, alpha; Colour c;} Bristle;
#define MAX_BRISTLES 256
Bristle bristle_start[MAX_BRISTLES]/*, *bristle_end = bristle_start*/;

const float brush_min_radius = 1.7, brush_max_radius = 12.0,
            brush_min_rate = 1.0,   brush_max_rate = 500.0;

void drawBrush(BITMAP *surface, Colour c, float radius, float rate, const int16_t *x, const int16_t *y, uint32_t points)
 {
  char first = 0;
  float lx = 0, ly = 0;
  float remainder = 0.0;
  uint32_t u, v;
  Bristle *b = bristle_start, *bristle_end;
  
  logMessage(Debug, "drawBrush invoked with %d points.", points);
  
  assert(isfinite(radius) && isfinite(rate));
  
  assert(surface);
  assert(x && y);
  
  if(rate < brush_min_rate) rate = brush_min_rate;
  else if(rate > brush_max_rate) rate = brush_max_rate;
  
  if(radius < brush_min_radius) radius = brush_min_radius;
  else if(radius > brush_max_radius) radius = brush_max_radius;
  
  /* We want an average of 1 per 2 pixels: */
  bristle_end = bristle_start + 1 +(int)(radius*3.1+fRNG());
  
  if(bristle_end > bristle_start + MAX_BRISTLES)
   bristle_end = bristle_start + MAX_BRISTLES;
  
  for(b = bristle_start; b < bristle_end; ++b)
   {
    b->c.r = c.r;
    b->c.g = c.g;
    b->c.b = c.b;
    b->x = b->lx = radius*2.0*((float)(b-bristle_start)/(float)(bristle_end-bristle_start)+fRNG()*0.2)-radius;
    b->y = b->ly = (fRNG()-0.5)*0.2*radius;
    b->remain = 0.0;
    b->alpha = 0.6+0.2*fRNG();
   }
  
  for(v = 0; v < points; ++v)
   {
    remainder = calcSplineChunk(x, y, points, v, 1.5, remainder);
    
    for(u = 0; u < spline_points; ++u)
     {
      if(first == 0)
       {
        lx = spline_x[u];
        ly = spline_y[u];
        first = 1;
        continue;
       }
      else
       {
        float vcos = ly-spline_y[u],
              vsin = spline_x[u]-lx,
              len = sqrt(SQR(vcos)+SQR(vsin));
        
        if(len < 0.0001)
         continue;
        
        vcos /= len;
        vsin /= len;
        
        lx = spline_x[u];
        ly = spline_y[u];
        
        if(first == 1)
         {
          for(b = bristle_start; b < bristle_end; ++b)
           {
            b->lx = spline_x[u]+b->x*vcos+b->y*vsin;
            b->ly = spline_y[u]+b->x*vsin-b->y*vcos;
           }
          
          first = 2;
          continue;
         }
        
        for(b = bristle_start; b < bristle_end; ++b)
         {
          Colour snag_colour;
          start_of_loop:
          
          assert(b >= bristle_start && b < bristle_end);
          
          if(getColour(surface, spline_x[u]+b->x, spline_y[u]+b->y, &snag_colour))
           {
            float snag = fRNG()*(fRNG()-b->alpha);
            
            if(snag > 0.01)
             {
              b->c.r = b->c.r*(1.0-snag)+snag_colour.r*snag;
              b->c.g = b->c.g*(1.0-snag)+snag_colour.g*snag;
              b->c.b = b->c.b*(1.0-snag)+snag_colour.b*snag;
             }
           }
          
          if(b->alpha > 0.0)
           {
            float alpha = b->alpha > 1.0 ? 1.0 : b->alpha, x, y;
            
            Colour hi, lo;
            
            hi.r = b->c.r+20;
            hi.g = b->c.g+20;
            hi.b = b->c.b+20;
            
            if(hi.r > 255.0) hi.r = 255.0;
            if(hi.g > 255.0) hi.g = 255.0;
            if(hi.b > 255.0) hi.b = 255.0;
            
            lo.r = b->c.r-16;
            lo.g = b->c.g-16;
            lo.b = b->c.b-16;
            
            if(lo.r < 0.0) lo.r = 0.0;
            if(lo.g < 0.0) lo.g = 0.0;
            if(lo.b < 0.0) lo.b = 0.0;
            
            x = spline_x[u]+b->x*vcos+b->y*vsin;
            y = spline_y[u]+b->x*vsin-b->y*vcos;
            
            drawLineRemainder(surface, b->lx+1.0, b->ly+1.0, x+1.0, y+1.0, lo, alpha, b->remain);
            drawLineRemainder(surface, b->lx-0.5, b->ly-0.5, x-0.5, y-0.5, hi, alpha, b->remain);
            drawLineRemainder(surface, b->lx+0.5, b->ly+0.5, x+0.5, y+0.5, b->c, alpha, b->remain);
            b->remain = drawLineRemainder(surface, b->lx-1.0, b->ly-1.0, x-1.0, y-1.0, b->c, alpha, b->remain);
            b->lx = x;
            b->ly = y;
           }
          else
           {
            if(--bristle_end == b)
             break;
            
            *b = *bristle_end;
            goto start_of_loop;
           }
          
          b->c.r = b->c.r*(1.0-b->alpha*b->alpha)+c.r*(b->alpha*b->alpha);
          b->c.g = b->c.g*(1.0-b->alpha*b->alpha)+c.g*(b->alpha*b->alpha);
          b->c.b = b->c.b*(1.0-b->alpha*b->alpha)+c.b*(b->alpha*b->alpha);
          
          b->alpha -= fRNG()*fRNG()/(rate+1.0)*0.6;
         }
       }
     }
   }
 }

void drawSmudgeCircle(BITMAP *surface, float radius, float rate, float x, float y, float dx, float dy)
 {
  float fy, temp;
  int cr, sr, u, v;
  Colour c;
  
  assert(isfinite(x) && isfinite(y));
  assert(isfinite(dx) && isfinite(dy));
  assert(isfinite(radius) && isfinite(rate));
  assert(radius > 0.0);
  assert(surface);
  
  fy = y-floor(y);
  cr = floor(radius+0.5);
  
  #define INST(X, Y) do{ \
     if(getColour(surface, (X)+dx, (Y)+dy, &c)) \
      alphaPixel(surface, (X), (Y), c, rate); \
     }while(0)
  
  if(dy < 0) for(v = 1-cr; v <= cr; ++v)
   {
    if((temp = radius*radius-SQR(v-fy)) <= 0)
     continue;
    
    sr = floor(sqrt(temp)+0.5);
    
    if(dx < 0) for(u = x-sr+1; u <= x+sr; ++u)
     INST(u, y+v);
    else for(u = x+sr; u > x-sr; --u)
     INST(u, y+v);
   }
  else for(v = cr; v >= 1-cr; --v)
   {
    if((temp = radius*radius-SQR(v-fy)) <= 0)
     continue;
    
    sr = floor(sqrt(temp)+0.5);
    
    if(dx < 0) for(u = x-sr+1; u <= x+sr; ++u)
     INST(u, y+v);
    else for(u = x+sr; u > x-sr; --u)
     INST(u, y+v);
   }
  
  #undef INST
 }

const float smudge_min_radius = 1.7, smudge_max_radius = 12.0,
            smudge_min_rate = 1.0,   smudge_max_rate   = 1000.0;

void drawSmudge(BITMAP *surface, float radius, float rate, const int16_t *x, const int16_t *y, uint32_t points)
 {
  float remainder = 0, last_x = 0, last_y = 0;
  uint32_t u, v;
  char first = 1;
  
  logMessage(Debug, "drawSmudge invoked with %d points.", points);
  
  assert(isfinite(radius) && isfinite(rate));
  
  assert(surface);
  assert(x && y);
  
  if(radius < smudge_min_radius) radius = smudge_min_radius;
  else if(radius > smudge_max_radius) radius = smudge_max_radius;
  
  if(rate < smudge_min_rate) rate = smudge_min_rate;
  else if(rate > smudge_max_rate) rate = smudge_max_rate;
  
  for(v = 0; v < points; ++v)
   {
    remainder = calcSplineChunk(x, y, points, v, 2.0, remainder);
    
    for(u = 0; u < spline_points; ++u)
     {
      if(first)
       {
        last_x = spline_x[u];
        last_y = spline_y[u];
        first = 0;
       }
      else
       {
        drawSmudgeCircle(surface, radius, rate/1200.0, spline_x[u], spline_y[u], last_x-spline_x[u], last_y-spline_y[u]);
        last_x = spline_x[u];
        last_y = spline_y[u];
       }
     }
   }
 }

const float pencil_min_radius = 1.5, pencil_max_radius = 8.0;

void drawPencil(BITMAP *surface, Colour c, float radius, const int16_t *x, const int16_t *y, uint32_t points)
 {
  float remainder = 0.0, area = 0;
  uint32_t u, v;
  
  logMessage(Debug, "drawPencil invoked with %d points.", points);
  
  assert(isfinite(radius));
  
  if(radius < pencil_min_radius) radius = pencil_min_radius;
  else if(radius > pencil_max_radius) radius = pencil_max_radius;
  
  for(v = 0; v < points; ++v)
   {
    remainder = calcSplineChunk(x, y, points, v, 1.0, remainder);
    
    for(u = 0; u < spline_points; ++u)
     {
      for(area = 0; area <= radius; area += 0.6)
       {
        float _x, _y, dist, offset=fRNG()*40.0-20.0;
        Colour t;
        t.r = c.r+offset;
        t.g = c.g+offset;
        t.b = c.b+offset;
        
        if(offset > 0)
         {
          if(t.r > 255.0) t.r = 255.0;
          if(t.g > 255.0) t.g = 255.0;
          if(t.b > 255.0) t.b = 255.0;
         }
        else
         {
          if(t.r < 0.0) t.r = 0.0;
          if(t.g < 0.0) t.g = 0.0;
          if(t.b < 0.0) t.b = 0.0;
         }
        
        do
         {
          _x = (fRNG()-0.5)*2.0*radius;
          _y = (fRNG()-0.5)*2.0*radius;
         } while((dist = sqrt(SQR(_x)+SQR(_y))) > radius);
        
        drawPoint(surface, spline_x[u]+_x, spline_y[u]+_y, t, 1.0);
       }
     }
   }
 }

const float stamp_min_radius = 1.5, stamp_max_radius = 8.0,
            stamp_min_rate = 0.0,   stamp_max_rate = 100.0;

void drawStamp(BITMAP *surface, Colour c, float radius, float rate, const int16_t *x, const int16_t *y, uint32_t points)
 {
  float remainder = 0.0;
  uint32_t u, v ;
  
  logMessage(Debug, "drawStamp invoked with %d points.", points);
  
  assert(isfinite(radius) && isfinite(rate));
  
  if(rate < stamp_min_rate) rate = stamp_min_rate;
  else if(rate > stamp_max_rate) rate = stamp_max_rate;
  
  if(radius < stamp_min_radius) radius = stamp_min_radius;
  else if(radius > stamp_max_radius) radius = stamp_max_radius;
  
  for(v=0; v < points; ++v)
   {
    remainder = calcSplineChunk(x, y, points, v, (radius*2.0)*(rate/100.0*9.0+1.0), remainder);
    
    for(u = 0; u < spline_points; ++u)
     drawCircle(surface, spline_x[u], spline_y[u], c, 1.0, radius);
   }
 }
